#!/usr/bin/env Rscript

#
#  This file is part of the `OmnipathR` R package
#
#  Copyright
#  2018-2021
#  Saez Lab, Uniklinik RWTH Aachen, Heidelberg University
#
#  File author(s): Alberto Valdeolivas
#                  Dénes Türei (turei.denes@gmail.com)
#                  Attila Gábor
#
#  Distributed under the MIT (Expat) License.
#  See accompanying file `LICENSE` or find a copy at
#      https://directory.fsf.org/wiki/License:Expat
#
#  Website: https://saezlab.github.io/omnipathr
#  Git repo: https://github.com/saezlab/OmnipathR
#


#' Signs for enzyme-substrate interactions
#'
#' Enzyme-substrate data does not contain sign (activation/inhibition), we
#' generate this information based on the interaction network.
#'
#' @param enzsub Enzyme-substrate data frame generated by
#'     \code{\link{import_omnipath_enzsub}}
#' @param interactions interaction data frame generated by
#'     \code{\link{import_omnipath_interactions}}
#'
#' @return Data frame of enzyme-substrate relationships with is_inhibition
#'     and is_stimulation columns.
#'
#' @export
#'
#' @examples
#' enzsub <- import_omnipath_enzsub(resources = c('PhosphoSite', 'SIGNOR'))
#' interactions <- import_omnipath_interactions()
#' enzsub <- get_signed_ptms(enzsub, interactions)
#'
#' @seealso \itemize{
#'     \item{\code{\link{import_omnipath_enzsub}}}
#'     \item{\code{\link{import_omnipath_interactions}}}
#' }
get_signed_ptms <- function(
    enzsub = import_omnipath_enzsub(),
    interactions = import_omnipath_interactions()
){

    signed.ptms <- merge(
        enzsub,
        interactions[,
            c("source", "target", "is_stimulation", "is_inhibition")
        ],
        by.x = c("enzyme", "substrate"),
        by.y = c("source", "target"),
        all.x = TRUE
    )

    return(signed.ptms)
}
